#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "ADE86.H"

static ARG86*	op;
static DIS86*	da;
static char*	p;
static char*	pmax;

char* reg8[] ={"AL","CL","DL","BL","AH","CH","DH","BH"};
char* reg16[]={"AX","CX","DX","BX","SP","BP","SI","DI"};
char* seg16[]={"ES","CS","SS","DS","FS","GS","??","??"};
char* rm16[] ={"BX+SI","BX+DI","BP+SI","BP+DI","SI","DI","BP","BX"};
char *jcc[] ={"O","NO","B","NB","E","NE","BE","NBE","S","NS","P",
		"NP","L","NL","LE","NLE"};

void
print_fmt(const char*format,...)
{
	va_list list;
	va_start(list,format);
	p+=_vsnprintf(p,pmax-p,format,list);
	va_end(list);
}

char*
print_const(Long _const, Byte size, bool _signed=true, bool eatsign=true)
{
	static char buf1[20],buf2[20],constneg=0;
	memset(buf1,0,sizeof(buf1));
	memset(buf2,0,sizeof(buf2));

	// print constant
	switch(size)
	{
	case(1):
		if(_const&0x80 && _signed) {
			sprintf(buf1,"0%02X",0xFF-_const+1);
			constneg=1;
		}
		else	sprintf(buf1,"0%02X",_const);
		break;
	case(2):
		if(_const&0x8000 && _signed) {
			sprintf(buf1,"0%04X",0xFFFF-_const+1);
			constneg=1;
		}
		else	sprintf(buf1,"0%04X",_const);
		break;
	case(4):
		if(_const&0x80000000 && _signed) {
			sprintf(buf1,"0%08X",0xFFFFFFFF-_const+1);
			constneg=1;
		}
		else	sprintf(buf1,"0%08X",_const);
		break;
	}

	// print any sign
	sprintf(buf2,constneg?"+%s":"-%s",
		isdigit(buf1[1]) ? &buf1[1] : &buf1[0]);

	if(eatsign)
		return &buf2[1];
	return buf2;
}

void
print_operand(Byte type, Byte size, Byte spec)
{
	Byte loc=Byte(spec&0xF0), mod, scale, base, index=0, disponly=1;

	// Decode RM as register or memory
	if(type==T_RM) {
		if((da->modrm&0xC0)==0xC0)
			type = T_REG;
		else	type = T_MEM;
	}

	// Set default operand size
	if(size==0 && (type==T_REL || type==T_OFFS)) {
		size = da->flags&C_USE16 ? Byte(2) : Byte(4);
		// 16bit mode
		if(da->flags&C_DS16) size = 2;
	}
	if(size==0 && (type==T_REG || type==T_MEM || type==T_IMM)) {
		size = da->flags&C_USE16 ? Byte(2) : Byte(4);
		// 16bit mode
		if(da->flags&C_DS16) size = 2;
		// Bit W
		if(da->flags&C_BITW && (da->opcode&1)==0) size = 1;
	}

	// Select register index
	if(loc==L_RG)
		index = Byte(da->opcode&7);	// opcode 00000xxx
	else if(loc==L_MRG)
		index = Byte((da->modrm>>3)&7);	// modrm  00xxx000
	else if(loc==L_MRM)
		index = Byte(da->modrm&7);	// modrm  00000xxx
	else	index = Byte(spec&7);		// Force specified

	// Print operand
	if(type==T_REG && size==1)
		print_fmt("%s",reg8[index]);
	else if(type==T_REG && size==2)
		print_fmt("%s",reg16[index]);
	else if(type==T_REG && size==4)
		print_fmt("E%s",reg16[index]);
	else if(type==T_SEG)
		print_fmt("%s",seg16[index]);
	else if(type==T_CRX)
		print_fmt("CR%d",index);
	else if(type==T_DRX)
		print_fmt("DR%d",index);
	else if(type==T_TRX)
		print_fmt("TR%d",index);
	else if(type==T_STX)
		print_fmt("ST%d",index);
	else if(type==T_MMX)
		print_fmt("MM%d",index);
	else if(type==T_REL)
		print_fmt("%s %s",size==1?"SHORT ":"NEAR ",
			print_const(da->address,4,false));
	else if(type==T_OFFS)
		print_fmt("[%s]",print_const(da->address,size,false));
	else if(type==T_IMM) {
		// spec = 0x01 Pesudo constant 1
		if(size==1 && spec&1) print_fmt("1");
		else	print_fmt("%s",print_const(da->immval,size,false));
	}
	else if(type==T_FAR) {
		print_fmt("%s:",print_const(da->selector,2,false));
		print_fmt("%s",print_const(da->address,2,false));
	}
	else if(type==T_MEM || type==T_FPTR) {
		if(type==T_FPTR) print_fmt("FAR ");
		mod = Byte(da->modrm&0xC0);
		base = Byte(da->modrm&7);
		// 16bit address
		if(da->flags&C_AS16) {
			if(mod==0 && base==6)
				print_fmt("[%s",print_const(da->address,2,false));
			else	print_fmt("[%s",rm16[base]);
			if(mod==0x40)
				print_fmt("%s",
					print_const(da->address,1,true,false));
			if(mod==0x80)
				print_fmt("%s",
					print_const(da->address,1,true,false));
			print_fmt("]");
			return;
		}
		//32bit address
		print_fmt("[");
		if(base!=4) {
			if(!(mod==0 && base==5)) {
				print_fmt("E%s",reg16[base]);
				disponly = 0;
			}
		} else {
			base = Byte(da->sib&7);
			index = Byte((da->sib>>3)&7);
			scale = Byte(da->sib>>6);
			if((mod==0 && base!=5) || mod!=0) {
				print_fmt("E%s",reg16[base]);
				disponly = 0;
				if(index!=4) print_fmt("+");
			}
			if(index!=4) {
				print_fmt("E%s",reg16[index]);
				disponly = 0; 
				if(scale!=0)
					print_fmt("*%s",
						scale==1 ? "2" :
						scale==2 ? "4" : "8");
			}
		}
		// disp
		if(mod==0 && base==5) {
			if(disponly)
				print_fmt("%s",
					print_const(da->address,4,false));
			else	print_fmt("%s",
					print_const(da->address,4,true,false));
		}

		if(mod==0x40)
			print_fmt("%s",print_const(da->address,1,true,false));
		if(mod==0x80)
			print_fmt("%s",print_const(da->address,4,true,false));
		print_fmt("]");
	}
}

int __cdecl
print_dis86(void* outptr, void* maxptr, DIS86* disasm, DBG86* debug)
{
	debug = 0;
	p = (char*)outptr;
	pmax = (char*)maxptr;
	da = disasm;
	op = (ARG86*)&da->tbl_entry->op1;
	// Instruction name
	if(da->flags&C_CC)
		print_fmt("%s%s ",da->tbl_entry->name,jcc[da->opcode&0xF]);
	else	print_fmt("%s ",da->tbl_entry->name);
	// "ENTER" instruction
	if(da->opcode==0xC8) {
		print_fmt("%s,",print_const(da->immval&0xFFFF,2,false));
		print_fmt("%s",print_const(da->immval>>16,1,false));
	}
	else for(int n=0,i=0; i<3; i++) {
		// Check operand exchange
		if(da->flags&C_BITD && da->opcode&2 && i<2)
			n = (n==0? 1 : 0);
		else	n = i;
		// Opernad separator
		if(i && op[i].optype!=0) print_fmt(",");
		// Print operand
		print_operand(Byte(op[n].optype&0xF0),
			Byte(op[n].optype&0xf),op[n].spec);
	}
	return p-(char*)outptr;
}
